# Make a sequence of CSS4 colors that are each as different as can be from the previous, given a palette of colors.
# See http://www.realtimerendering.com/blog/low-discrepancy-color-sequences/
# Eric Haines, erich@acm.org
# Version 1.0, October 31, 2022
use strict;
use warnings;
use POSIX qw/ceil/;
$| = 1;	# flush text to screen immediately

my $findcolors = 139;	# how many colors to find.
my $colorcube = 1;		# 1 - use linear RGB for distances; 0 - use CIELAB

my $outputsize = 32;	# size of color tiles output to PPM file
my $xres = 10;			# number of colors per row
my $yres = ceil($findcolors / $xres);	# no touch!

# output file names
my $fileroot = "css4";
my $imgfilename = $fileroot . '_colorcube' . $colorcube . '.ppm';
my $txtfilename = $fileroot . '_colorcube' . $colorcube . '.txt';

#------------------------
my @cssnames = (
"aliceblue",
"antiquewhite",
"aqua",
"aquamarine",
"azure",
"beige",
"bisque",
"black",
"blanchedalmond",
"blue",
"blueviolet",
"brown",
"burlywood",
"cadetblue",
"chartreuse",
"chocolate",
"coral",
"cornflowerblue",
"cornsilk",
"crimson",
"cyan",
"darkblue",
"darkcyan",
"darkgoldenrod",
"darkgray",
"darkgreen",
"darkgrey",
"darkkhaki",
"darkmagenta",
"darkolivegreen",
"darkorange",
"darkorchid",
"darkred",
"darksalmon",
"darkseagreen",
"darkslateblue",
"darkslategray",
"darkslategrey",
"darkturquoise",
"darkviolet", 
"deeppink",
"deepskyblue",
"dimgray",
"dimgrey",
"dodgerblue", 
"firebrick",
"floralwhite",
"forestgreen",
"fuchsia",
"gainsboro",
"ghostwhite", 
"gold",
"goldenrod",
"gray",
"green",
"greenyellow",
"grey",
"honeydew",
"hotpink",
"indianred",
"indigo",
"ivory",
"khaki",
"lavender",
"lavenderblush",
"lawngreen",
"lemonchiffon",
"lightblue",
"lightcoral", 
"lightcyan",
"lightgoldenrodyellow",
"lightgray",
"lightgreen",
"lightgrey",
"lightpink",
"lightsalmon",
"lightseagreen",
"lightskyblue",
"lightslategray",
"lightslategrey",
"lightsteelblue",
"lightyellow",
"lime",
"limegreen",  
"linen",
"magenta",
"maroon",
"mediumaquamarine", 
"mediumblue", 
"mediumorchid",
"mediumpurple",
"mediumseagreen",
"mediumslateblue",
"mediumspringgreen",
"mediumturquoise",
"mediumvioletred",
"midnightblue",
"mintcream",
"mistyrose",
"moccasin",
"navajowhite",
"navy",
"oldlace",
"olive",
"olivedrab",
"orange",
"orangered",
"orchid",
"palegoldenrod",
"palegreen",
"paleturquoise",
"palevioletred",
"papayawhip", 
"peachpuff",
"peru",
"pink",
"plum",
"powderblue", 
"purple",
"rebeccapurple",
"red",
"rosybrown",
"royalblue",  
"saddlebrown",
"salmon",
"sandybrown", 
"seagreen",
"seashell",
"sienna",
"silver",
"skyblue",
"slateblue",
"slategray",
"slategrey",
"snow",
"springgreen",
"steelblue",  
"tan",
"teal",
"thistle",
"tomato",
"turquoise",  
"violet",
"wheat",
"white",
"whitesmoke", 
"yellow",
"yellowgreen");

my @csscolors = (
[240, 248, 255],
[250, 235, 215],
[0, 255, 255],
[127, 255, 212],
[240, 255, 255],
[245, 245, 220],
[255, 228, 196],
[0, 0, 0],
[255, 235, 205],
[0, 0, 255],
[138, 43, 226],
[165, 42, 42],
[222, 184, 135],
[95, 158, 160],
[127, 255, 0],
[210, 105, 30],
[255, 127, 80],
[100, 149, 237],
[255, 248, 220],
[220, 20, 60],
[0, 255, 255],
[0, 0, 139],
[0, 139, 139],
[184, 134, 11],
[169, 169, 169],
[0, 100, 0],
[169, 169, 169],
[189, 183, 107],
[139, 0, 139],
[85, 107, 47],
[255, 140, 0],
[153, 50, 204],
[139, 0, 0],
[233, 150, 122],
[143, 188, 143],
[72, 61, 139],
[47, 79, 79],
[47, 79, 79],
[0, 206, 209],
[148, 0, 211],
[255, 20, 147],
[0, 191, 255],
[105, 105, 105],
[105, 105, 105],
[30, 144, 255],
[178, 34, 34],
[255, 250, 240],
[34, 139, 34],
[255, 0, 255],
[220, 220, 220],
[248, 248, 255],
[255, 215, 0],
[218, 165, 32],
[128, 128, 128],
[0, 128, 0],
[173, 255, 47],
[128, 128, 128],
[240, 255, 240],
[255, 105, 180],
[205, 92, 92],
[75, 0, 130],
[255, 255, 240],
[240, 230, 140],
[230, 230, 250],
[255, 240, 245],
[124, 252, 0],
[255, 250, 205],
[173, 216, 230],
[240, 128, 128],
[224, 255, 255],
[250, 250, 210],
[211, 211, 211],
[144, 238, 144],
[211, 211, 211],
[255, 182, 193],
[255, 160, 122],
[32, 178, 170],
[135, 206, 250],
[119, 136, 153],
[119, 136, 153],
[176, 196, 222],
[255, 255, 224],
[0, 255, 0],
[50, 205, 50],
[250, 240, 230],
[255, 0, 255],
[128, 0, 0],
[102, 205, 170],
[0, 0, 205],
[186, 85, 211],
[147, 112, 219],
[60, 179, 113],
[123, 104, 238],
[0, 250, 154],
[72, 209, 204],
[199, 21, 133],
[25, 25, 112],
[245, 255, 250],
[255, 228, 225],
[255, 228, 181],
[255, 222, 173],
[0, 0, 128],
[253, 245, 230],
[128, 128, 0],
[107, 142, 35],
[255, 165, 0],
[255, 69, 0],
[218, 112, 214],
[238, 232, 170],
[152, 251, 152],
[175, 238, 238],
[219, 112, 147],
[255, 239, 213],
[255, 218, 185],
[205, 133, 63],
[255, 192, 203],
[221, 160, 221],
[176, 224, 230],
[128, 0, 128],
[102, 51, 153], # CSS4 rebeccapurple
[255, 0, 0],
[188, 143, 143],
[65, 105, 225],
[139, 69, 19],
[250, 128, 114],
[244, 164, 96],
[46, 139, 87],
[255, 245, 238],
[160, 82, 45],
[192, 192, 192],
[135, 206, 235],
[106, 90, 205],
[112, 128, 144],
[112, 128, 144],
[255, 250, 250],
[0, 255, 127],
[70, 130, 180],
[210, 180, 140],
[0, 128, 128],
[216, 191, 216],
[255, 99, 71],
[64, 224, 208],
[238, 130, 238],
[245, 222, 179],
[255, 255, 255],
[245, 245, 245],
[255, 255, 0],
[154, 205, 50] );

my $tablelength = scalar(@csscolors);

my @csslinearcolors;

my @cssused = (0) x $tablelength;


#seed color, to start
my @reds;
my @greens;
my @blues;
my @idxseq;

# white:
$reds[0] = $greens[0] = $blues[0] = 1;
$idxseq[0] = 144;

open(TFH, '>', $txtfilename) or die $!;

print "$txtfilename\n";
print TFH "$txtfilename\n";

my $i;
my $j;

my $deleted = 0;
# look for duplicate colors and merge names
for ($i=0;$i<$tablelength-1;$i++) {
	for ($j=$i+1;$j<$tablelength;$j++) {
		if ($csscolors[$i][0] == $csscolors[$j][0] && $csscolors[$i][1] == $csscolors[$j][1] && $csscolors[$i][2] == $csscolors[$j][2] ) {
			# we don't actually merge, but do 
			print "WARNING: colors $cssnames[$i] and $cssnames[$j] have same triplet ($csscolors[$i][0],$csscolors[$i][1],$csscolors[$i][2]) - merging entries.\n";
			$cssnames[$i] .= '/' . $cssnames[$j];
			$cssnames[$j] = 'DELETED' . $j;
			$deleted++;
		}
	}
}

if ( $findcolors > $tablelength - $deleted ) {
	print "WARNING: findcolors $findcolors > tablelength $tablelength - deleted $deleted. Adjusting $findcolors.\n";
	$findcolors = $tablelength - $deleted;
}

# CIE LAB perceptual space differences
for ($i=0;$i<$tablelength;$i++) {
	($csslinearcolors[$i][0], $csslinearcolors[$i][1], $csslinearcolors[$i][2] ) =
		$colorcube ? 
		# measure using linear RGB values
		&sRGBtoLinearRGB($csscolors[$i][0],$csscolors[$i][1],$csscolors[$i][2]) :
		# CIE LAB perceptual space differences
		&sRGBtoCIELAB($csscolors[$i][0],$csscolors[$i][1],$csscolors[$i][2]);
	
	if ( 1 ) {
		# print out triplets
		printf "For %s, sRGB is (%d, %d, %d) converts to CIE-L*ab (%0.2f, %0.2f, %0.2f)\n", $cssnames[$i], $csscolors[$i][0],$csscolors[$i][1],$csscolors[$i][2], $csslinearcolors[$i][0], $csslinearcolors[$i][1], $csslinearcolors[$i][2];
	}
}

my ($veryclosestneighbor, $closestneighbor);
for ($i=0;$i<$findcolors;$i++) {
	my $farthest = -1;

	my ($r,$g,$b);
	if ( $i > 0 ) {
		$idxseq[$i] = -1;
		#print "color $i\n";
		# look at every color in table for the best next color. Find its closest approach distance to previous colors chosen.
		for (my $tidx=0;$tidx<$tablelength;$tidx++) {
			if ( $cssused[$tidx] == 0 ) {
				# color not yet used
				my $distance = &colorDistance($csslinearcolors[$tidx][0],$csslinearcolors[$tidx][1],$csslinearcolors[$tidx][2],$i);
				# We want to take the largest closest distance.
				if ($distance > $farthest) {
					$farthest = $distance;
					$idxseq[$i] = $tidx;
					$veryclosestneighbor = $closestneighbor;
				}
			}
		}
	}
	my $idx = $idxseq[$i];
	$cssused[$idx] = $i+1;	# or just true would be fine
	if ( $idx < 0 ) {
		print "ERROR: exit on $i color - no match found. Perhaps there are matching colors in your table?\n";
		exit 0;
	}
	$reds[$i] = $csslinearcolors[$idx][0];
	$greens[$i] = $csslinearcolors[$idx][1];
	$blues[$i] = $csslinearcolors[$idx][2];
	if ( $i == 0 ) {
		printf "Color %3d is (%3d, %3d, %3d) and is named %s\n",
			$i,
			$csscolors[$idx][0], $csscolors[$idx][1], $csscolors[$idx][2], $cssnames[$idx];
		printf TFH "Color %3d is (%3d, %3d, %3d) and is named %s\n",
			$i,
			$csscolors[$idx][0], $csscolors[$idx][1], $csscolors[$idx][2], $cssnames[$idx];
	} else {
		printf "Color %3d is (%3d, %3d, %3d) and is named %s, nearest neighbor is %s at a distance of %0.4f\n",
			$i,
			$csscolors[$idx][0], $csscolors[$idx][1], $csscolors[$idx][2], $cssnames[$idx], $cssnames[$idxseq[$veryclosestneighbor]], sqrt($farthest);
		printf TFH "Color %3d is (%3d, %3d, %3d) and is named %s, nearest neighbor is %s at a distance of %0.4f\n",
			$i,
			$csscolors[$idx][0], $csscolors[$idx][1], $csscolors[$idx][2], $cssnames[$idx], $cssnames[$idxseq[$veryclosestneighbor]], sqrt($farthest);
	}
}


open(IFH, '>', $imgfilename) or die $!;

print IFH "P3\n";

printf IFH "%d %d\n255\n", $xres*$outputsize, $yres*$outputsize;

my $colornum = 0;
for (my $y=0;$y<$yres;$y++) {
	for (my $yl = 0; $yl < $outputsize; $yl++){
		$colornum = $y * $xres;
		for (my $x=0;$x<$xres;$x++) {
			# is color to be output?
			if ( $colornum < $findcolors ) {
				my $ci = $idxseq[$colornum];
				my $xloc = $x * $outputsize;
				my $yloc = int($colornum / $xres) * $outputsize;
				my $loc = $x * $outputsize + $yloc * $outputsize * $xres * $outputsize;

				my $color = sprintf "%d %d %d ", $csscolors[$ci][0], $csscolors[$ci][1], $csscolors[$ci][2];
				for (my $xl = 0; $xl < $outputsize; $xl++){
					#$img[$loc+$yl*$xres*$outputsize+$xl] = $color;
					print IFH $color;
				}
				$colornum++;
			} else {
				# no color output
				for (my $xl = 0; $xl < $outputsize; $xl++){
					print IFH "0 0 0 ";
				}
			}
		}
		printf IFH "\n";
	}
}

close IFH;

#===========================
	
# returns distance of -1 if color is rejected for some reason
sub colorDistance {
	my ($r,$g,$b,$cidx) = @_;

	my $max_distance = 99999;
	# go through previous colors and find closest one
	for (my $c=0;$c<$cidx;$c++){
		my $dist = ($r-$reds[$c])*($r-$reds[$c]) + ($g-$greens[$c])*($g-$greens[$c]) + ($b-$blues[$c])*($b-$blues[$c]);

#print "Dist is $dist for $r,$g,$b vs. $c: $reds[$c], $greens[$c], $blues[$c]\n";
		if ($dist < $max_distance) {
			# The max_distance for the test point is the minimum of how far it is from all previous colors
			$max_distance = $dist;
			if ( $dist == 0 ) {
				# means we've already used this color, so it definitely won't be The One - early out
				return -1;
			}
			$closestneighbor = $c;
#print "max dist is now $max_distance\n";
		}
	}
	return $max_distance;
}

sub sRGBtoLinearRGB {
	# Code snippets from http://www.easyrgb.com/en/math.php
	#sR, sG and sB (Standard RGB) input range = 0 ÷ 255
	#X, Y and Z output refer to a D65/2° standard illuminant.
	my ($sR,$sG,$sB) = @_;

	my $var_R = ( $sR / 255 );
	my $var_G = ( $sG / 255 );
	my $var_B = ( $sB / 255 );

	# convert from sRGB to linear values
	if ( $var_R > 0.04045 ) { $var_R = ( ( $var_R + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_R = $var_R / 12.92; }
	if ( $var_G > 0.04045 ) { $var_G = ( ( $var_G + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_G = $var_G / 12.92; }
	if ( $var_B > 0.04045 ) { $var_B = ( ( $var_B + 0.055 ) / 1.055 ) ** 2.4; }
	else                   { $var_B = $var_B / 12.92; }
	
	return ($var_R, $var_G, $var_B);
}

sub sRGBtoCIELAB {
	# https://github.com/NVlabs/flip/blob/main/cpp/CPP/color.h
	#sR, sG and sB (Standard RGB) input range = 0 ÷ 255
	#X, Y and Z output refer to a D65/2° standard illuminant.
	my ($sR,$sG,$sB) = @_;

	my ($var_R, $var_G, $var_B) = &sRGBtoLinearRGB($sR,$sG,$sB);

	# Source: https://www.image-engineering.de/library/technotes/958-how-to-convert-between-srgb-and-ciexyz
	# Assumes D65 standard illuminant.
	my $a11 = 10135552.0 / 24577794.0;
	my $a12 = 8788810.0 / 24577794.0;
	my $a13 = 4435075.0 / 24577794.0;
	my $a21 = 2613072.0 / 12288897.0;
	my $a22 = 8788810.0 / 12288897.0;
	my $a23 = 887015.0 / 12288897.0;
	my $a31 = 1425312.0 / 73733382.0;
	my $a32 = 8788810.0 / 73733382.0;
	my $a33 = 70074185.0 / 73733382.0;

	my $X = $a11 * $var_R + $a12 * $var_G + $a13 * $var_B;
	my $Y = $a21 * $var_R + $a22 * $var_G + $a23 * $var_B;
	my $Z = $a31 * $var_R + $a32 * $var_G + $a33 * $var_B;

	my $delta = 6.0 / 29.0;
	my $deltaSquare = $delta * $delta;
	my $deltaCube = $delta * $deltaSquare;
	my $factor = 1.0 / (3.0 * $deltaSquare);
	my $term = 4.0 / 29.0;

	# The default illuminant is D65.
	my @Reference = (1.052156925, 1.000000000, 0.918357670);
	my $var_X = $X / $Reference[0];
	my $var_Y = $Y / $Reference[1];
	my $var_Z = $Z / $Reference[2];

	$var_X = ($var_X > $deltaCube ? $var_X**(1.0 / 3.0) : $factor * $var_X + $term);
	$var_Y = ($var_Y > $deltaCube ? $var_Y**(1.0 / 3.0) : $factor * $var_Y + $term);
	$var_Z = ($var_Z > $deltaCube ? $var_Z**(1.0 / 3.0) : $factor * $var_Z + $term);
	my $L = 116.0 * $var_Y - 16.0;
	my $a = 500.0 * ($var_X - $var_Y);
	my $b = 200.0 * ($var_Y - $var_Z);
	
	return ($L, $a, $b);
}
